#include "LibSip_SIPTypesAndValues.hh"
#include "LibSip_Common.hh"

#include "sip_codec_message_body.hh"
#include "loggers.hh"

#include "converter.hh"

int sip_codec_message_body::encode (const LibSip__MessageBodyTypes::MessageBody& msg, OCTETSTRING& data)
{
  loggers::get_instance().log_msg(">>> sip_codec_message_body::encode: ", msg);

  //  TRACE_INITIALIZE(TRACE_LEVEL7, NULL);
  //  OSIP_TRACE (osip_trace (__FILE__, __LINE__, OSIP_ERROR, NULL, ">>> sip_codec_request::decode\n"));

  if (msg.ischosen(LibSip__MessageBodyTypes::MessageBody::ALT_sdpMessageBody)) {
    const LibSip__SDPTypes::SDP__Message& sdp = msg.sdpMessageBody();
    sdp_message_t* sdp_body = nullptr;
    if (encode_message_body_sdp(sdp, &sdp_body) == -1) {
      loggers::get_instance().warning("sip_codec_message_body::encode: Failed to encode SDP message");
      return -1;
    }
    char* buff = nullptr;
    ::sdp_message_to_str(sdp_body, &buff);
    if (buff != nullptr) {
      data = char2oct(CHARSTRING(buff));
      osip_free(buff);
      sdp_message_free(sdp_body);
    }
  } else {
    loggers::get_instance().warning("sip_codec_message_body::encode: Unsupported variant");
    return -1;
  }

  loggers::get_instance().log_msg("<<< sip_codec_message_body::encode: Failed to encode SDP message: ", data);
  return 0;
}

int sip_codec_message_body::decode (const OCTETSTRING& data, LibSip__MessageBodyTypes::MessageBody& msg, params* p_params)
{
  loggers::get_instance().log(">>> sip_codec_message_body::decode");

  //  TRACE_INITIALIZE(TRACE_LEVEL7, NULL);
  //  OSIP_TRACE (osip_trace (__FILE__, __LINE__, OSIP_ERROR, NULL, ">>> sip_codec_request::decode\n"));
  loggers::get_instance().error("sip_codec_message_body::decode: Not implemented yet");
  return -1;
}

int sip_codec_message_body::encode_message_body(const LibSip__MessageBodyTypes::MessageBody& p_message_body, osip_message_t* p_sip_message)
{ // TODO Rename into encode_message
  loggers::get_instance().log(">>> sip_codec_message_body::encode_message_body");

  osip_content_length* content_length = ::osip_message_get_content_length(p_sip_message);
  loggers::get_instance().log("sip_codec_message_body::encode_message_body: content_length= %s", content_length->value);
  if (p_message_body.ischosen(LibSip__MessageBodyTypes::MessageBody::ALT_sdpMessageBody)) {
    const LibSip__SDPTypes::SDP__Message& sdp = p_message_body.sdpMessageBody();
    sdp_message_t* sdp_body = nullptr;
    if (encode_message_body_sdp(sdp, &sdp_body) == -1) {
      loggers::get_instance().warning("sip_codec_message_body::encode_message_body: Failed to decode SDP message");
      return -1;
    }
    char* buff = nullptr;
    ::sdp_message_to_str(sdp_body, &buff);
    if (buff != nullptr) {
      ::osip_message_set_body(p_sip_message, buff, strlen(buff));
      osip_free(buff);
      sdp_message_free(sdp_body);
    }
  }
  
  return 0;
} // End of method encode_message_body

int sip_codec_message_body::encode_message_body_sdp(const LibSip__SDPTypes::SDP__Message& p_sdp_message, sdp_message_t** p_sdp_body) {
  loggers::get_instance().log_msg(">>> sip_codec_message_body::encode_message_body_sdp: ", p_sdp_message);

  ::sdp_message_init(p_sdp_body);

  // Encode mandatory fields
  // Version
  INTEGER version = p_sdp_message.protocol__version();
  ::sdp_message_v_version_set(*p_sdp_body, (char*)::strdup(std::to_string(static_cast<const int>(version)).c_str()));
  LibSip__SDPTypes::SDP__Origin origin = p_sdp_message.origin();
  // Origin
  loggers::get_instance().log_msg("sip_codec_message_body::encode_message_body_sdp: origin: ", origin);
  ::sdp_message_o_origin_set(
                             *p_sdp_body,
                             (char*)::strdup(static_cast<const char*>(origin.user__name())),
                             (char*)::strdup(static_cast<const char*>(origin.session__id())),
                             (char*)::strdup(static_cast<const char*>(origin.session__version())),
                             (char*)::strdup(static_cast<const char*>(origin.net__type())),
                             (char*)::strdup(static_cast<const char*>(origin.addr__type())),
                             (char*)::strdup(static_cast<const char*>(origin.addr()))
                             );
  // Session name
  ::sdp_message_s_name_set(*p_sdp_body, (char*)::strdup(static_cast<const char*>(p_sdp_message.session__name())));
  // Time, t
  ::sdp_message_s_name_set(*p_sdp_body, (char*)::strdup(static_cast<const char*>(p_sdp_message.session__name())));
  loggers::get_instance().log_msg("sip_codec_message_body::encode_message_body_sdp: times: ", p_sdp_message.times());
  for (int i = 0; i < p_sdp_message.times().lengthof(); i++) {
    const LibSip__SDPTypes::SDP__time& time = p_sdp_message.times()[i];
    loggers::get_instance().log_msg("sip_codec_message_body::encode_message_body_sdp: time: ", time);
    ::sdp_message_t_time_descr_add(*p_sdp_body, (char*)::strdup(static_cast<const char*>(time.time__field().start__time())), (char*)::strdup(static_cast<const char*>(time.time__field().stop__time())));
    if (time.time__repeat().is_present()) {
      const LibSip__SDPTypes::SDP__repeat__list& l = static_cast<const OPTIONAL<LibSip__SDPTypes::SDP__repeat__list>&>(*time.time__repeat().get_opt_value());
      loggers::get_instance().log_msg("sip_codec_message_body::encode_message_body_sdp: l: ", l);
      for (int j = 0; j < l.lengthof(); j++) {
        const LibSip__SDPTypes::SDP__repeat& r = l[j];
        // TODO            r.repeat__interval();
        // TODO            r.active();
        for (int k = 0; k < r.offsets().lengthof(); k++) {
      	const LibSip__SDPTypes::SDP__typed__time& s = r.offsets()[k];
      	// TODO              ::sdp_message_r_repeat_add(p_sdp_body, k, char *value);
        } // End of 'for' statement
        // TODO        ::sdp_message_r_repeat_add(p_sdp_body, int pos_time_descr, char *value)
      } // End of 'for' statement
    }
  } // End of 'for' statement

  // Encode optional fields
  // email, e
  loggers::get_instance().log_msg("sip_codec_message_body::encode_message_body_sdp: emails: ", p_sdp_message.emails());
  if (p_sdp_message.emails().is_present()) {
    const LibSip__SDPTypes::SDP__email__list& l = static_cast<const LibSip__SDPTypes::SDP__email__list&>(*p_sdp_message.emails().get_opt_value());
    for (int i = 0; i < l.lengthof(); i++) {
      const LibSip__SDPTypes::SDP__contact& e = l[i];
      std::string str;
      if (e.disp__name().is_present()) {
        str = static_cast<const char*>(static_cast<const CHARSTRING&>(*e.disp__name().get_opt_value()));
        str += " ";
      }
      str += static_cast<const char*>(e.addr__or__phone());
      ::sdp_message_e_email_add(*p_sdp_body, ::strdup(str.c_str()));
    } // End of 'for' statement
  }
  // Information, i
  loggers::get_instance().log_msg("sip_codec_message_body::encode_message_body_sdp: information: ", p_sdp_message.information());
  if (p_sdp_message.information().is_present()) {
    ::sdp_message_i_info_set(*p_sdp_body, 0, ::strdup(static_cast<const CHARSTRING&>(*p_sdp_message.information().get_opt_value())));
  }
  // Media list, m
  loggers::get_instance().log_msg("sip_codec_message_body::encode_message_body_sdp: medias: ", p_sdp_message.media__list());
  if (p_sdp_message.media__list().is_present()) {
    const LibSip__SDPTypes::SDP__media__desc__list& l = static_cast<const LibSip__SDPTypes::SDP__media__desc__list&>(*p_sdp_message.media__list().get_opt_value());
    int i = 0;
    for ( ; i < l.lengthof(); i++) {
      const LibSip__SDPTypes::SDP__media__desc m = l[i];
      loggers::get_instance().log_msg("sip_codec_message_body::encode_message_body_sdp: media: ", m);
      if (m.media__field().ports().num__of__ports().is_present()) {
        int i = static_cast<const INTEGER&>(*m.media__field().ports().num__of__ports().get_opt_value());
        ::sdp_message_m_media_add(*p_sdp_body, (char*)::strdup(static_cast<const char*>(m.media__field().media())), (char*)::strdup(std::to_string(m.media__field().ports().port__number()).c_str()), (char*)::strdup(std::to_string(i).c_str()), (char*)::strdup(static_cast<const char*>(m.media__field().transport())));
      } else {
        ::sdp_message_m_media_add(*p_sdp_body, (char*)::strdup(static_cast<const char*>(m.media__field().media())), (char*)::strdup(std::to_string(m.media__field().ports().port__number()).c_str()), NULL, (char*)::strdup(static_cast<const char*>(m.media__field().transport())));
      }
    } // End of 'for' statement
    ::sdp_message_endof_media(*p_sdp_body, i);
  }
  // Attribute list, a
  loggers::get_instance().log_msg("sip_codec_message_body::encode_message_body_sdp: attributes: ", p_sdp_message.attributes());
  if (p_sdp_message.attributes().is_present()) {
    const LibSip__SDPTypes::SDP__attribute__list& l = static_cast<const LibSip__SDPTypes::SDP__attribute__list&>(*p_sdp_message.attributes().get_opt_value());
    int i = 0;
    for ( ; i < l.lengthof(); i++) {
      const LibSip__SDPTypes::SDP__attribute a = l[i];
      loggers::get_instance().log_msg("sip_codec_message_body::encode_message_body_sdp: attribute: ", a);
      if (a.ischosen(LibSip__SDPTypes::SDP__attribute::ALT_rtpmap)) {
        std::string str("rtpmap:");
        str += static_cast<const char*>(a.rtpmap().codec().encoding());
        str += a.rtpmap().codec().clockrate();
        if (a.rtpmap().codec().parameters().is_present()) {
          str += "/";
          str += static_cast<const char*>(static_cast<const CHARSTRING&>(*a.rtpmap().codec().parameters().get_opt_value()));
        }
        ::sdp_message_a_attribute_add(*p_sdp_body, 0, (char*)::strdup(static_cast<const char*>(a.rtpmap().payload__type())), (char*)::strdup(str.c_str()));
      } else if (a.ischosen(LibSip__SDPTypes::SDP__attribute::ALT_maxptime)) {
        std::string str("maxptime:");
        str += static_cast<const char*>(a.maxptime().attr__value());
        ::sdp_message_a_attribute_add(*p_sdp_body, 0, (char*)::strdup(str.c_str()), NULL);
      } else {
        loggers::get_instance().warning("sip_codec_message_body::encode_message_body_sdp: attribute not processed");
      }
    } // End of 'for' statement
  }
  // TODO To be continued

  loggers::get_instance().log("<<< sip_codec_message_body::encode_message_body_sdp");
  return 0;
}

void sip_codec_message_body::decode_message_body(const osip_message_t* p_sip_message, OPTIONAL<LibSip__MessageBodyTypes::MessageBody>& p_message_body)
{ // TODO Rename into decode_message
  loggers::get_instance().log(">>> sip_codec_message_body::decode_message_body");

  osip_content_length* content_length = ::osip_message_get_content_length(p_sip_message);
  loggers::get_instance().log("sip_codec_message_body::decode_message_body: content_length= %s", content_length->value);
  if (strcmp((const char*)content_length->value, "0") == 0) {
    p_message_body.set_to_omit();
    return;
  }
  // TODO Align the body the content_length

  osip_content_type_t* content_type = ::osip_message_get_content_type(p_sip_message);
  osip_body_t* body;
  int result = ::osip_message_get_body(p_sip_message, 0, &body);
  loggers::get_instance().log("sip_codec_message_body::decode_message_body: result=%d", result);
  if (result != 0) {
    p_message_body.set_to_omit();
    return;
  }
  if (content_type != nullptr) {
    loggers::get_instance().log("sip_codec_message_body::decode_message_body: content-type.type: %s", content_type->type);
    loggers::get_instance().log("sip_codec_message_body::decode_message_body: content-type.subtype: %s", content_type->subtype);
    LibSip__MessageBodyTypes::MessageBody msg_body;
    if (std::string(content_type->subtype).compare("sdp") == 0) {
      LibSip__SDPTypes::SDP__Message sdp_header;
      sdp_message_t* sdp;
      ::sdp_message_init(&sdp);
      if (::sdp_message_parse(sdp, body->body) == 0) {
        // Mandatroy fields
        // Version, v
        char *hvalue = ::sdp_message_v_version_get(sdp);
        if (hvalue != nullptr) {
          sdp_header.protocol__version() = std::stoi(hvalue);
        }
        // Origin, o
        sdp_header.origin() = LibSip__SDPTypes::SDP__Origin(
                                                            ::sdp_message_o_username_get(sdp),
                                                            ::sdp_message_o_sess_version_get(sdp),
                                                            ::sdp_message_o_sess_version_get(sdp),
                                                            ::sdp_message_o_nettype_get(sdp),
                                                            ::sdp_message_o_addrtype_get(sdp),
                                                            ::sdp_message_o_addr_get(sdp)
                                                            );
        // Session name, s
        if ((hvalue = ::sdp_message_s_name_get(sdp)) != nullptr) {
          sdp_header.session__name() = CHARSTRING(hvalue);
        }
        // Time, t
        int i = 0;
        if ((hvalue = ::sdp_message_t_start_time_get(sdp, i)) != nullptr) {
          LibSip__SDPTypes::SDP__time__list l;
          do {
            LibSip__SDPTypes::SDP__time t;
            LibSip__SDPTypes::SDP__time__field tf;
            tf.start__time() = CHARSTRING(hvalue);
            tf.stop__time() = CHARSTRING(sdp_message_t_stop_time_get(sdp, i));
            t.time__field() = tf;
            t.time__repeat().set_to_omit(); // TODO
            l[i] = t;
            i += 1;
          } while ((hvalue = ::sdp_message_t_start_time_get(sdp, i)) != nullptr);
          sdp_header.times() = l;
        } else {
          loggers::get_instance().warning("sip_codec_message_body::decode_message_body: No start time in sdp");
        }
        // Optional fields
        // Emails, e
        i = 0;
        if ((hvalue = ::sdp_message_e_email_get(sdp, i)) != nullptr) {
          LibSip__SDPTypes::SDP__email__list l;
          do {
            LibSip__SDPTypes::SDP__contact e;
            e.addr__or__phone() = CHARSTRING(hvalue);
            e.disp__name().set_to_omit(); // TODO To be refined
            l[i] = e;
            i += 1;
          } while ((hvalue = ::sdp_message_e_email_get(sdp, i)) != nullptr);
          sdp_header.emails() = OPTIONAL<LibSip__SDPTypes::SDP__email__list>(l);
        } else {
          loggers::get_instance().warning("sip_codec_message_body::decode_message_body: No emails in sdp");
          sdp_header.emails().set_to_omit();
        }
        // Information, i
        i = 0;
        if ((hvalue = ::sdp_message_i_info_get(sdp, i)) != nullptr) {
          sdp_header.information() = OPTIONAL<CHARSTRING>(CHARSTRING(hvalue));
        } else {
          loggers::get_instance().warning("sip_codec_message_body::decode_message_body: No information in sdp");
          sdp_header.information().set_to_omit();
        }
        // Media, m
        i = 0;
        if ((hvalue = ::sdp_message_m_media_get(sdp, i)) != nullptr) {
          LibSip__SDPTypes::SDP__media__desc__list l;
          do {
            loggers::get_instance().log("sip_codec_message_body::decode_message_body: Media='%s'", hvalue);
            // Media description
            LibSip__SDPTypes::SDP__media__field f;
            f.media() = CHARSTRING(hvalue);
            LibSip__SDPTypes::SDP__media__port p;
            hvalue = ::sdp_message_m_port_get(sdp, i);
            p.port__number() = std::stoi(hvalue);
            if ((hvalue = ::sdp_message_m_number_of_port_get(sdp, i)) == nullptr) {
              p.num__of__ports().set_to_omit();
            } else {
              p.num__of__ports() = OPTIONAL<INTEGER>(std::stoi(hvalue));
            }
            f.ports() = p;
            f.transport() = CHARSTRING(::sdp_message_m_proto_get(sdp, i));
            LibSip__SDPTypes::SDP__fmt__list fl;
            int j = 0;
            while ((hvalue = ::sdp_message_m_payload_get(sdp, i, j)) != nullptr) {
              fl[j++] = CHARSTRING(hvalue);
            } // End of 'for' statement
            f.fmts() = fl;
            LibSip__SDPTypes::SDP__media__desc m;
            loggers::get_instance().log_msg("sip_codec_message_body::decode_message_body: m=", m);
            m.media__field() = f;
            // TODO
            // Media attributes
            m.attributes().set_to_omit();
            m.bandwidth().set_to_omit();
            m.information().set_to_omit();
            m.connections().set_to_omit();
            m.key().set_to_omit();
            l[i] = m;
            i += 1;
          } while ((hvalue = ::sdp_message_m_media_get(sdp, i)) != nullptr);
          sdp_header.media__list() = OPTIONAL<LibSip__SDPTypes::SDP__media__desc__list>(l);
        } else {
          loggers::get_instance().warning("sip_codec_message_body::decode_message_body: No media in sdp");
          sdp_header.media__list().set_to_omit();
        }
        // Attributes, a
        i = 0;
        sdp_attribute_t* a;
        if ((a = ::sdp_message_attribute_get(sdp, 0, i)) != nullptr) {
          LibSip__SDPTypes::SDP__attribute__list l;
          int j = 0;
          do {
            loggers::get_instance().log("sip_codec_message_body::decode_message_body: Attributes#%d: '%s'='%s'", i, a->a_att_field, a->a_att_value);
            std::vector<std::string> output = converter::get_instance().split(a->a_att_value, " ");
            // Attribute description
            LibSip__SDPTypes::SDP__attribute f;
            if (strcmp(a->a_att_field, "rtpmap") == 0) {
              LibSip__SDPTypes::SDP__attribute__rtpmap__codec c;
              std::vector<std::string> o = converter::get_instance().split(a->a_att_value, "/");
              c.encoding() = CHARSTRING(o[0].c_str());
              c.clockrate() = CHARSTRING(o[1].c_str());
              if (o.size() == 3) {
                c.parameters() = OPTIONAL<CHARSTRING>(CHARSTRING(o[2].c_str()));
              } else {
                c.parameters().set_to_omit();
              }
              f.rtpmap() = LibSip__SDPTypes::SDP__attribute__rtpmap(CHARSTRING(output[0].c_str()), c);
            } else if (strcmp(a->a_att_field, "maxptime") == 0) {
              LibSip__SDPTypes::SDP__attribute__maxptime m;
              f.maxptime() = LibSip__SDPTypes::SDP__attribute__maxptime(CHARSTRING((const char*)a->a_att_value));
            } else {
              loggers::get_instance().warning("sip_codec_message_body::decode_message_body: Unprocessed attributes#%d: '%s'='%s'", i, a->a_att_field, a->a_att_value);
            }
            //loggers::get_instance().log_msg("sip_codec_message_body::decode_message_body: f: ", f);
            if (f.is_bound()) {
              l[j++] = f;
            }
            i += 1;
          } while ((a = ::sdp_message_attribute_get(sdp, 0, i)) != nullptr);
          if (j == 0) { // Empty list
            sdp_header.attributes().set_to_omit();
          } else {
            sdp_header.attributes() = OPTIONAL<LibSip__SDPTypes::SDP__attribute__list>(l);
          }
        } else {
          loggers::get_instance().warning("sip_codec_message_body::decode_message_body: No attributes in sdp");
          sdp_header.attributes().set_to_omit();
        }
        
        // TODO
        sdp_header.uri().set_to_omit();
        sdp_header.phone__numbers().set_to_omit();
        sdp_header.connection().set_to_omit();
        // Bandwidth
        //        sdp_bandwidth_t* b = ::sdp_message_bandwidth_get(sdp, i, j);
        sdp_header.bandwidth().set_to_omit();

        sdp_header.timezone__adjustments().set_to_omit();
        sdp_header.key().set_to_omit();

        // Set message body
        msg_body.sdpMessageBody() = sdp_header;
        loggers::get_instance().log_msg("sip_codec_message_body::decode_message_body: sdp:", msg_body);
      } else {
        loggers::get_instance().warning("sip_codec_message_body::decode_message_body: sdp_message_parse failed, fallback to textplain");
        msg_body.textplain() = CHARSTRING(body->body);
      }
    } else {
      loggers::get_instance().warning("sip_codec_message_body::decode_message_body: Unsuported content-type, fallback to textplain");
      msg_body.textplain() = CHARSTRING(body->body);
    }
    p_message_body = OPTIONAL<LibSip__MessageBodyTypes::MessageBody>(msg_body);
  } else if ((body != nullptr) && (body->body != nullptr)) {
    LibSip__MessageBodyTypes::MessageBody msg_body;
    msg_body.textplain() = CHARSTRING(body->body);
    p_message_body = OPTIONAL<LibSip__MessageBodyTypes::MessageBody>(msg_body);
  } else {
    p_message_body.set_to_omit();
  }
  
  loggers::get_instance().log_msg("<<< sip_codec_message_body::decode_message_body: ", p_message_body);
} // End of method decode_message_body
